/***************************************************************************
* Copyright (C)1999 The XAOL Project, All Rights Reserved.                 *
* WWW: http://www.xaol.org/                                                *
***************************************************************************/
// Modules.c

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "Modules.h"
#include "AOLSession.h"
#include "AOLModules.h"

// Linked list of modules loaded
Module* modules = NULL;
int numModules = 0;

Module* MNew () {
	Module* module = NULL;
	
	module = (Module*) MAlloc(sizeof(Module));
	if (module != NULL) {
		snprintf(module->name, 256, "Unnamed Module");
		module->type = Mod_Uninit;
		memset((void*) module->messages, 0, sizeof(module->messages));	
		module->data = NULL;
		module->id = numModules++;
		module->next = NULL;
		module->previous = NULL;
		module->checkValue = 128;
	 	}
	
	return (module);
	}
	
void MDispose (Module* module) {
	if (MIsValid(module))
		MFree((void*) module);
	}

int MIsValid (Module* module) {
	return (module != NULL && module->checkValue == 128);
	}
		
XAOLError MAdd (Module* module) {
	Module* end = modules;

	if (! MIsValid(module))
		return (ME_InvalidModule);
		
	if (end == NULL) {
		module->next = NULL;
		module->previous = NULL;
		modules = module;
		}
	else {
		while (end->next != NULL)
			end = end->next;
	
		end->next = module;
		module->previous = end;
		module->next = NULL;
		}
		
	return (No_Error);
	}
	
XAOLError MRemove (Module* module) {
	Module* n, *p;
	
	if (MIsValid(module)) {
		if (module->previous == NULL)
			return (ME_ModuleNotFound);
			
		p = module->previous;
		n = module->next;
		module->previous = NULL;
		module->next = NULL;
		p->next = n;
		}
	else
		return (ME_InvalidModule);
	
	return (No_Error);
	}
	
XAOLError MDispatchMessage (int messageType, void* messageData) {
	Module* module = modules;
	XAOLError error;
	
	while (MIsValid(module)) {
		if (module->messages[messageType] == 1) {
			error = MSendMessage(module, messageType, messageData);
			if (error != No_Error)
				return (error);
			}
			
		module = module->next;
		}
	
	return (No_Error);
	}
	
int MNumModules () {
	return (numModules);
	}
	
XAOLError MGetModuleByID (int id, Module** module) {
	Module* m = modules;
	
	if (MIsValid(m)) {
		while (m->id != id) {
			if (! MIsValid(m))
				return (ME_ModuleNotFound);
			m = m->next;
			}
		
		*module = m;
		}
	else
		return (ME_InvalidModule);
		
	return (No_Error);
	}
	
XAOLError MGetModuleByName (char* name, Module** module) {
	Module* m = modules;
	
	if (MIsValid(m)) {
		// TODO - CodeWarrior Seems to yell about strcasecmp
		while (strcmp(name, m->name) != 0) {
			if (! MIsValid(m))
				return (ME_ModuleNotFound);
			m = m->next;
			}
		
		*module = m;
		}
	else
		return (ME_InvalidModule);
		
	return (No_Error);
	}

XAOLError MGetModuleByKind (ModType type, Module** module) {
	Module* m = modules;
	
	if (MIsValid(m)) {
		// TODO - CodeWarrior Seems to yell about strcasecmp
		while (m->type != type) {
			if (! MIsValid(m))
				return (ME_ModuleNotFound);
			m = m->next;
			}
		
		*module = m;
		}
	else
		return (ME_InvalidModule);
		
	return (No_Error);
	}
	
XAOLError MLoad (Module** module, char* path) {
	Module* m = NULL;
	XAOLError error;
	ModInitExitPtr addr;
	
	m = MNew();
	if (m == NULL) {
		XAOL_DEBUG_Print_("Error making module ref\n");
		return (Mem_Error);
		}
		
	error = MPrivateLoad (m, path);
	if (error != No_Error) {
		XAOL_DEBUG_Print_("Error in MPrivateLoad\n");
		MDispose(m);
		return (error);
		}

	error = MGetSymbol (m, kModInit, (void**) &addr);
	if (error != No_Error) {
		XAOL_DEBUG_Print_("Error retreiving module init symbol\n");
		MDispose(m);
		return (error);
		}
	
	error = addr(m);
	if (error != No_Error) {
		XAOL_DEBUG_Print_("Error initializing the module.\n");
		MDispose(m);
		return (error);
		}

	error = MAdd(m);
	if (error != No_Error) {
		XAOL_DEBUG_Print_("Error adding the module to the list.\n");
		MDispose(m);
		return (error);
		}
		
	*module = m;
	return (No_Error);
	}
	
XAOLError MUnload (Module* module) {
	XAOLError error;
	ModInitExitPtr addr;
	
	if (MIsValid(module)) {
		error = MRemove(module);
		if (error != No_Error)
			return (error);
		
		error = MGetSymbol (module, kModExit, (void**) &addr);
		if (error == No_Error) {	
			error = addr(module);
			if (error != No_Error)
				XAOL_DEBUG_Print_("Error unloading the module.\n");
			}
			
		error = MPrivateUnload(module);
		if (error != No_Error)
			return (error);
		
		MDispose(module);
		}
	else
		return (ME_InvalidModule);
	
	return (No_Error);
	}

XAOLError MStartModules (AOLSession* session) {
	Module* module = modules;
	ModStartStopPtr addr;
	XAOLError error;
	
	if (! MIsValid(module)) 
		return (No_Error);
		
	while (module != NULL) {
		error = MGetSymbol(module, kModStart, (void**) &addr);
		if (error == No_Error) {
			error = addr(module, session);
			if (error != No_Error)
				return (error);
			}
			
		module = module->next;
		}

	return (No_Error);
	}
